#include <ctype.h>

#include "ball.h"

char      *progname;
GtkWidget *main_window;
GtkWidget *darea;
GdkPixmap *backing_pixmap = NULL;

char *errMemory   = "%s: error--cannot allocate memory\n";
char *errPalette  = "%s: error--cannot allocate color palette\n";

static gint ball_radius = DEFAULT_BALL_RADIUS;

static int parse_options(int argc, char **argv);



/*
 * Signal and event handlers
 */
static void darea_destroy(GtkWidget *widget)
{
    gdk_pixmap_unref(backing_pixmap);

    return;
}

static gint darea_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
    /* Redraw clobbered area of backing pixmap */
    gdk_draw_pixmap(widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    backing_pixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);

    return FALSE;
}

/*
 * In configure event we (re)create the backing pixmap with new size.
 */
static gint darea_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
    gint R, width, height;

    if ( backing_pixmap )
        gdk_pixmap_unref(backing_pixmap);

    width  = widget->allocation.width;
    height = widget->allocation.height;

    backing_pixmap = gdk_pixmap_new(widget->window, width, height, -1);

    /* Blank out the rectangle */
    gdk_draw_rectangle( backing_pixmap, widget->style->black_gc,
                        TRUE, 0, 0, width, height);

    R = width < height ? (width-1)/2 : (height-1)/2;

    /* Draw the ball */
    draw_ball( (GdkDrawable *) backing_pixmap, palette,
               num_shades, Ka, Kd, Ks,
               width/2, height/2, R );

    return TRUE;
}


int main(int argc, char **argv)
{
    int        i;
    GtkWidget *main_box, *widget;

    progname = *argv;

    /* GTK initialization */
    gtk_init(&argc, &argv);
    /* Process command line arguments */
    i = parse_options(argc, argv);
    if ( i == 0 )
        return 1;
    else if ( i == 1 )
        return 0;

    /* Main window */
    main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title( GTK_WINDOW(main_window), "Ball" );
    /* Can shrink and grow */
    gtk_window_set_policy( GTK_WINDOW(main_window), TRUE, TRUE, FALSE );
    /* Will quit on destroy signal event */
    gtk_signal_connect( GTK_OBJECT(main_window), "destroy",
                        GTK_SIGNAL_FUNC(gtk_main_quit), NULL );
    /* Pack main vbox into main window (homogeneous, spacing) */
    main_box = gtk_vbox_new(FALSE, 0);
    gtk_container_add ( GTK_CONTAINER(main_window), main_box );
    /* Pack drawing area into main box */
    darea = gtk_drawing_area_new();
    gtk_widget_set_usize( darea, ball_radius, ball_radius );
    /* (..., expand, fill, padding)*/
    gtk_box_pack_start ( GTK_BOX(main_box), darea, TRUE, TRUE, 0 );
    gtk_signal_connect( GTK_OBJECT(darea), "expose_event",
                        GTK_SIGNAL_FUNC(darea_expose_event), NULL );
    gtk_signal_connect( GTK_OBJECT(darea), "configure_event",
                        GTK_SIGNAL_FUNC(darea_configure_event), NULL );
    gtk_signal_connect( GTK_OBJECT(darea), "destroy",
                        GTK_SIGNAL_FUNC(darea_destroy), NULL );
    /* Pack controls box into main box (this will aslo setup palette, that's why
     * we realize main window) */
    gtk_widget_realize(main_window);
    widget = create_controls_box();
    if ( !widget )
        return 1;
    gtk_box_pack_start ( GTK_BOX(main_box), widget, FALSE, FALSE, 0 );

    /* Pop up the whole shebang */
    gtk_widget_show(main_box);
    gtk_widget_show(darea);
    gtk_widget_show(main_window);

    gtk_main();

    return 0;
}


static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
static char *errBallRadius    = "%s: error--ball's radius must be a positive integer\n";
static char *errColorIntensity= "%s: error--color component must be an integer in range [0, 100]\n";
static char *errNumShades     = "%s: error--number of shades must be an integer in range [2, 256]\n";
static char *errKaKdKs        = "%s: error--Ka, Kd, and Ks must be integers in range [0, shades-1]\n";
static char *errCoeffSum      = "%s: error--the sum of Ka, Kd, and Ks must not exceed (shades-1)\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument (%s)\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\n\
Usage: %s [-h] [-w radius] [-r red] [-g green] [-b blue]\n\
         [-n shades] [-a Ka] [-d Kd] [-s Ks]\n\
\n\
-h              display this help message\n\
-w radius       ball's width (radius)                         (default: %3d)\n\
-r red          ball color's red component   (range: [0, 100]; default: %3d)\n\
-g red          ball color's green component (range: [0, 100]; default: %3d)\n\
-b blue         ball color's blue component  (range: [0, 100]; default: %3d)\n\
-n shades       number of shades of the color specified by -r, -g, and -b above\n\
                to allocate in the palette   (range: [2, 256]; default: %3d)\n\
-a Ka           Ka coefficient (dispersed light proportion)\n\
                                        (range: [0, shades-1]; default: %3d)\n\
-d Kd           Kd coefficient (diffused light proportion)\n\
                                        (range: [0, shades-1]; default: %3d)\n\
-s Ks           Ks coefficient (ball surface's luster)\n\
                                        (range: [0, shades-1]; default: %3d)\n\
\n\
Note: the sum of Ka, Kd, and Ks must not exceed (shades-1)\n\
\n";

/* Return codes:
 * 0    - error, exit(1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
static int parse_options(int argc, char **argv)
{
    char c, *ptr;
    gint  i;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":w:r:g:b:n:a:d:s:h")) != -1 )
        switch ( c ) {
        case 'w':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 0 ) {
                fprintf(stderr, errBallRadius, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            ball_radius = i;
            break;
        case 'r': case 'g': case 'b':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 0 || i > 100 ) {
                fprintf(stderr, errColorIntensity, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            switch( c ) {
            case 'r':
                red = i;
                break;
            case 'g':
                green = i;
                break;
            case 'b':
                blue = i;
                break;
            default:
                assert(0); /* should never happen */
                break;
            }
            break;
        case 'n':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 2 || i > 256 ) {
                fprintf(stderr, errNumShades, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            num_shades = i;
            break;
        case 'a': case 'd': case 's':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errKaKdKs, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            switch( c ) {
            case 'a':
                Ka = i;
                break;
            case 'd':
                Kd = i;
                break;
            case 's':
                Ks = i;
                break;
            default:
                assert(0); /* should never happen */
                break;
            }
            break;
        case 'h':
            printf(usage, progname, DEFAULT_BALL_RADIUS,
                   DEFAULT_RED, DEFAULT_GREEN, DEFAULT_BLUE,
                   DEFAULT_NUM_SHADES, DEFAULT_KA, DEFAULT_KD, DEFAULT_KS);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            return 0;
            break;
        case ':': { /* missing option argument */
            char *msg = NULL;

            switch ( optopt ) {
            case 'w':
                msg = "ball's radius";
                break;
            case 'r':
                msg = "ball color's red component";
                break;
            case 'g':
                msg = "ball color's green component";
                break;
            case 'b':
                msg = "ball color's blue component";
                break;
            case 'n':
                msg = "number of shades";
                break;
            case 'a':
                msg = "dispersed light proportion";
                break;
            case 'd':
                msg = "diffused light proportion";
                break;
            case 's':
                msg = "ball's surface luster";
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
            fprintf(stderr, errArgumentReq, progname, optopt, msg);
            fprintf(stderr, usehelp, progname);
            return 0;
            break;
        }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }

    if ( Ka < 0 || Ka >= num_shades ||
         Kd < 0 || Kd >= num_shades ||
         Ks < 0 || Ks >= num_shades ) {
        fprintf(stderr, errKaKdKs, progname);
        fprintf(stderr, usehelp, progname);
        return 0;
    }

    if ( Ka + Kd + Ks >= num_shades ) {
        fprintf(stderr, errCoeffSum, progname);
        fprintf(stderr, usehelp, progname);
        return 0;
    }

    return -1;
}



/*
 * Local variables:
 * compile-command: "make -k && ./ball"
 * End:
 */
