#include "convex_hull.h"
#include <limits.h>     /* DBL_MAX */
#include <math.h>       /* sqrt() */

/*
 * Triangulates polygon described by points in points list.  Points must be
 * listed clockwise.  Last point in the list should not close with first,
 * i.e. last point should not be duplicated at the end of the list.
 *
 * Returns list of triangles on success, NULL on error (points list does not
 * describe a valid polygon)
 */
list *convex_hull(list *points)
{
    list  *hull = new list;
    list  *remaining;
    pitem *initial, *v, *ptr;
    /* To clean up the mess some, vector names are uppercase, point names are
     * lowercase */
    int    UV_x, UV_y, u_x, u_y;
    double UV_len, UV_NORM_x, UV_NORM_y, A_x, A_y;
    double max_prod, prod;

    if ( !points || points->count() < 2 )
        return NULL;

    /* Make a scrap list of points */
    remaining = new list;
    remaining = points->copy();

    /* Find the leftmost (and if more than one--the lowest of those) point */
    initial = (pitem *) remaining->head();
    ptr = (pitem *) initial->next;
    while ( ptr ) {
        if ( ptr->p().x < initial->p().x                                 ||
             ptr->p().x == initial->p().x && ptr->p().y < initial->p().y    )
            initial = ptr;
        ptr = (pitem *) ptr->next;
    }

    u_x = initial->p().x;
    u_y = initial->p().y;
    A_x = 0;
    A_y = 1;
    while ( 1 ) {
        /* Find point v such that angle between vector UV and vector A is
         * smallest.  This is equivalent to finding vector UV which, when
         * normalized, gives the largest scalar product with vector A (we keep A
         * normalized throughout) */
        max_prod = -DBL_MAX; /* smallest possible */
        ptr = (pitem *) remaining->head();
        v = NULL; /* pointer to the point v */
        while ( ptr ) {
            if ( ptr->p().x != u_x || ptr->p().y != u_y ) { /* skip point u */
                /* Calculate parameters of vector UV */
                UV_x = ptr->p().x - u_x;
                UV_y = ptr->p().y - u_y;
                UV_len = sqrt(UV_x * UV_x + UV_y * UV_y);
                /* Normalize vector UV */
                UV_NORM_x = UV_x / UV_len;
                UV_NORM_y = UV_y / UV_len;
                /* Find the scalar product and check if it is larger than
                 * previous */
                prod = UV_NORM_x * A_x + UV_NORM_y * A_y;
                if ( prod > max_prod ) {
                    max_prod = prod;
                    v = ptr;
                }
            }
            ptr = (pitem *) ptr->next;
        }
        assert(v);
        /* Add the point we've found to the hull list */
        hull->add(new pitem(v->p()));
        /* Bail out as soon as we have added the initial point */
        if ( v == initial )
            break;
        /* Now recalculate normalized vector UV, were V is the point we have
         * found.  Although we have already calculated this before, it may be
         * more expensive to remember normalized coordinates in the main loop
         * each time we reestablish V than just to recalculate it here */
        UV_x = v->p().x - u_x;
        UV_y = v->p().y - u_y;
        UV_len = sqrt(UV_x * UV_x + UV_y * UV_y);
        /* Take point V and normalized vector UV as point U and vector A for the
         * next iteration */
        A_x = UV_x / UV_len;
        A_y = UV_y / UV_len;
        u_x = v->p().x;
        u_y = v->p().y;
        /* Remove point V--we will no longer consider it as it is already in the
         * hull */
        remaining->del(v);
    }

    delete remaining; /* this was a scrap list */

    return hull;
}



/*
 * Local variables:
 * compile-command: "make -k && ./convex_hull -d < points.set"
 * End:
 */
