#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>
#include <unistd.h>

#include <vga.h>
#include <vgakeyboard.h>


static char  *progname;
static int    scrwidth, scrheight, colordepth;

static int parse_options(int argc, char **argv);


static char *errMemory   = "%s: error--cannot allocate memory\n";
static char *errKeyboard = "%s: error--failed to initialize keyboard to raw \
mode\n";

#define MAX_LOOPS_COUNT         10
#define DEFAULT_LOOPS_COUNT     4
static int loops = DEFAULT_LOOPS_COUNT; /* number of loops for the wiggle */

static void draw_wiggle(int cx, int cy, int radius, int length, int loops);

void line(unsigned char *screen, unsigned char color,
          int x, int y, int X, int Y);



int main(int argc, char **argv)
{
    int  i;
    int *palvec, r, g, b;

    progname = *argv;

    /* Process command line arguments */
    i = parse_options(argc, argv);
    if ( i == 0 )
        return -1;
    else if ( i == 1 )
        return 0;

    /* Arranging custom palette--we only use 240 slots because 240 is a divisor
     * of 3600, thus making the colors close when the wiggle is drawn (see
     * draw_wiggle)
     */
    palvec = malloc( sizeof(int) * 3 * 240 );
    if ( !palvec ) {
        fprintf(stderr, errMemory, *argv);
        return 1;
    }

    for ( i = 0; i < 60; i++ ) {   /* BLACK -> RED */
        *(palvec + 3*i)     = i;   /*  0 -> 59 */
        *(palvec + 3*i + 1) = 0;   /*     0    */
        *(palvec + 3*i + 2) = 0;   /*     0    */
    }
    for ( i = 60; i < 120; i++ ) {  /* RED -> GREEN */
        *(palvec + 3*i)     = 119-i;/* 59 -> 0  */
        *(palvec + 3*i + 1) = i-60; /*  0 -> 59 */
        *(palvec + 3*i + 2) = 0;    /*     0    */
    }
    for ( i = 120; i < 180; i++ ) { /* GREEN -> BLUE */
        *(palvec + 3*i)     = 0;    /*     0    */
        *(palvec + 3*i + 1) = 179-i;/* 59 -> 0  */
        *(palvec + 3*i + 2) = i-120;/*  0 -> 59 */
    }
    for ( i = 180; i < 240; i++ ) {/* BLUE->BLACK, and meet you at BLACK->RED */
        *(palvec + 3*i)     = 0;    /*     0    */
        *(palvec + 3*i + 1) = 0;    /*     0    */
        *(palvec + 3*i + 2) = 239-i;/* 59 -> 0  */
    }

    /* Setup graphics */
    vga_init();
    vga_setmode(G320x200x256);

    scrwidth   = vga_getxdim();
    scrheight  = vga_getydim();
    colordepth = vga_getcolors();

    /* We'll loop until user hits a key, so need to enter raw mode */
    if ( keyboard_init() ) {
        vga_setmode(TEXT);
        fprintf(stderr, errKeyboard, *argv);
        free((void *) palvec);
        return 1;
    }

    /* Set our custom palette */
    vga_waitretrace();
    /* We leave slot 0 alone so that background is never changed */
    vga_setpalvec(1, 240, palvec);
    /* Draw the thingy */
    draw_wiggle(160, 100, 50, 35, loops-1);

    /* Enter the loop */
    while ( 1 ) {
        if ( keyboard_update() ) /* user pressed a key */
            break;
        /* Remember first color */
        r = *(palvec + 0);
        g = *(palvec + 1);
        b = *(palvec + 2);
        /* Rotate the palette 1 slot left */
        for ( i = 1; i < 240; i++ ) {
            *(palvec + i*3 - 3) = *(palvec + i*3);
            *(palvec + i*3 - 2) = *(palvec + i*3 + 1);
            *(palvec + i*3 - 1) = *(palvec + i*3 + 2);
        }
        *(palvec + 239*3)     = r;
        *(palvec + 239*3 + 1) = g;
        *(palvec + 239*3 + 2) = b;
        /* Update palette */
        vga_waitretrace();
        vga_setpalvec(1, 240, palvec);
    }

    /* Shut down the whole shebang */
    keyboard_close();
    vga_setmode(TEXT);
    free((void *) palvec);

    return 0;
}


/* loops is actually one less then number of loops for the wiggle */
static void draw_wiggle(int cx, int cy, int radius, int length, int loops)
{
    int    color;
    int    i;
    long   j;
    int    rx, ry;
    double a;

    color = 5;
    for ( i = 0, j = 0; i < 3600; i++, j+=loops ) {
        rx = radius * cos(i*M_PI*2/3600) + cx;
        ry = radius * sin(i*M_PI*2/3600) + cy;
        a = j * 2 * M_PI / 3600;
        //vga_waitretrace();
        line( vga_getgraphmem(), color/5,
              rx - length * cos(a), ry + length * sin(a),
              rx + length * cos(a), ry - length * sin(a) );
        color++;
        if ( color >= 241*5 )
            color = 5;
        //getchar();
    }

    return;
}


static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character \
`\\x%x'\n";
static char *errLoopsRange    = "%s: error--number of loops must be an integer \
in range from 1 to %d\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument \
(%s)\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\
Usage: %s [-h] [-l loops_count]\n\
\n\
-h              display this help message\n\
-l loops_count  the wiggle will have 2*loops_count loops, loops_count can be\n\
                from 1 to %d (default: %d)\n\
\n";

/* Return codes:
 * 0    - error, exit(-1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
static int parse_options(int argc, char **argv)
{
    char c, *ptr;
    int  l;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":l:h")) != -1 )
        switch ( c ) {
         case 'l':
            l = (int) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || l < 1 || l > MAX_LOOPS_COUNT ) {
                fprintf(stderr, errLoopsRange, *argv, MAX_LOOPS_COUNT);
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            loops = l;
            break;
        case 'h':
            printf(usage, *argv, MAX_LOOPS_COUNT, DEFAULT_LOOPS_COUNT);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, *argv, optopt);
                fprintf(stderr, usehelp, *argv);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, *argv, optopt);
                fprintf(stderr, usehelp, *argv);
            }
            return 0;
            break;
        case ':': /* missing option argument */
            switch ( optopt ) {
            case 'l':
                fprintf(stderr, errArgumentReq, *argv, optopt, "number of loops");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }

    return -1;
}



/*
 * Local variables:
 * compile-command: "make -k"
 * End:
 */
