#! /usr/bin/python

import sys
import getopt
from math import *

#
#  Quadrants:
#
#       -----------
#      |     |     |
#      |x1y1 |x0y0 |
#      |-----|-----|
#      |     |     |
#      |x2y2 |x3y3 |
#  (X,Y)-----------
#
#
#  Directions:     Orientation:              O | D | Starting Quadrant
#                                           ---|---|-------------------
#      1                                     1 | 0 | 2
#      |           1 - counterclockwise      1 | 1 | 3
#  2 <- -> 0                                 1 | 2 | 0
#      |           0 - clockwise             1 | 3 | 1
#      3                                     0 | 0 | 1
#                                            0 | 1 | 2
#                                            0 | 2 | 3
#                                            0 | 3 | 0
#
def hilbert(X, Y, A, DIR, ORIENT):
    a          = A / 2
    # Coordinates of quadrants
    x, y       = [0, 0, 0, 0], [0, 0, 0, 0]
    x[0], y[0] = X + a, Y + a
    x[1], y[1] = X,     Y + a
    x[2], y[2] = X    , Y
    x[3], y[3] = X + a, Y
    # Direction, starting quadrant and increment value in each quadrant
    if ORIENT == 1:   # counterclockwise
        dir = [ (DIR + 1) % 4, DIR, DIR, (DIR + 3) % 4 ]
        quadrant = (DIR + 2) % 4
        q = 1
    else:             # clockwise
        dir = [ (DIR + 3) % 4, DIR, DIR, (DIR + 1) % 4 ]
        quadrant = (DIR + 1) % 4
        q = 3 # after MOD 4 will become -1
    # Orientation in each quadrant
    orient = [ not ORIENT, ORIENT, ORIENT, not ORIENT ]

    if a == 1:
        for i in range(4): # draw three sides of a 2x2 square
            print 1, x[quadrant], y[quadrant]
            quadrant = (quadrant + q) % 4
    else:
        for i in range(4): # recursively draw all four quadrants in order
            hilbert(x[quadrant], y[quadrant], a, dir[i], orient[i])
            quadrant = (quadrant + q) % 4



# Default values
draw_moore = 0
color      = 7
side       = 4
label      = "Hilbert/Moore Curve (order %d)"
label_h    = "Hilbert Curve (order %d)"
label_m    = "Moore Curve (order %d)"
use_default_label = 1

# Error messages
errTryHelp          = "Try `" + sys.argv[0] + " -h' for help\n"
errIntegerReq       = "Option `%s' requires an integer argument or a script evaluating to such\n"
errSideRange        = "Square side parameter must be a positive integer\n"


usage = \
"Usage: " + sys.argv[0] + """ [-h] [-s side] [-c color] [-l label] [-C]n

-h        show this help message
-m        draw moore curve (default: draw hilbert curve)
-s side   integer parameter for square side (square side == 2**side)
-c color  integer color to use for graph (default: """ + str(color) + """)
-l label  use specified label (default: `""" + label + """')
"""



# Process command line arguments
try:
    optionlist, args = getopt.getopt(sys.argv[1:], "ms:c:l:h")
except getopt.GetoptError, (msg, opt):
    sys.stderr.write(msg + "\n")
    sys.stderr.write(errTryHelp)
    sys.exit(1)

for opt, val in optionlist:
    try:
        if opt == "-c":
            color = int(eval(val))
        if opt == "-s":
            side = int(val)
            if side <= 0:
                sys.stderr.write(errSideRange)
                sys.stderr.write(errTryHelp)
                sys.exit(1)
    except (ValueError, SyntaxError, NameError):
        sys.stderr.write(errIntegerReq % (opt))
        sys.stderr.write(errTryHelp)
        sys.exit(1)

    if opt == "-l":
        label = val
        use_default_label = 0
    if opt == "-h":
        print usage
        sys.exit(0)
    if opt == "-m":
        draw_moore = 1

if use_default_label:
    if not draw_moore:
        label = label_h % side
    else:
        label = label_m % side

# Output label
print label

# Output color and real-world viewport coordinates
len = 2**side
print color, 0, 0, len - 1, len - 1

if not draw_moore or len == 2:
    print 0, 0, 0 # "move-to"
    hilbert(0, 0, len, 1, 0)
else:
    len /= 2
    print 0, len - 1, 0 # "move-to"
    hilbert(  0,   0, len, 2, 0)
    hilbert(  0, len, len, 2, 0)
    hilbert(len, len, len, 0, 0)
    hilbert(len,   0, len, 0, 0)
