#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#include "wrapper.h"

#define AXES_COLOR      1
#define LABEL_LEN       256


char *errUnknownOption = "%s: error--unknown option `-%c'\n";
char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
char *errInteger       = "%s: error--%s must be an integer\n";
char *errArgumentReq   = "%s: error--option `%s' requires an argument (%s)\n";
char *errDataFileReq   = "%s: error--you must specify at least one data file\n";
char *errUnexpectedEOF = "%s: error--unexpected end of file `%s'\n";
char *errParsingFile   = "%s: error parsing file `%s'\n";
char *errFont          = "%s: error--cannot load font\n";
char *errOpenFile      = "%s: error--cannot open file `%s'\n";
char *errLabelTooLong  = "\
%s: error parsing file `%s'--length\n\
of the label should not exceed %d characters";

char *usehelp = "Try `%s -h' for help\n";
char *usage   = "\
Usage: %s [-h] [-x xmin] [-y ymin] [-X xmax] [-Y ymax]\n\
         [-A] [-a axes_color] [-L] [-s] [-d nsec] data_file [...]\n\n\
-h       display this help message\n\
-x xmin  x-coord of ll corner of viewport (default: 0);\n\
-y ymin  y-coord of ll corner of viewport (default: 0);\n\
-X xmax  x-coord of ur corner of viewport (default: scrwidth - 1);\n\
-Y ymax  y-coord of ur corner of viewport (default: scrheight - 1);\n\
-a       color (integer) to use for axes (default: %d)\n\
-A       suppress axes, ignore -a (default: show axes)\n\
-L       suppress graph labels (default: show graph labels)\n\
-s       stretch graph in both directions, breaking aspect ratio\n\
         (default: preserve aspect ratio)\n\
-d nsec  delay each stroke nsec nanoseconds (default: 0)\n\n\
Note: real-world viewport coordinates are ignored in all but\n\
      the first data files\n\
      if you leave out the coordinates, the whole screen will be\n\
      used as a viewport\n\
      if you give negative coordinates, they will be set to defaults\n\n";


char  *progname;
int    scrwidth, scrheight, colordepth;
char   label[LABEL_LEN];
int    stretch, showaxes, axescolor, showlabels; /* options */
int    xvirt, yvirt, Xvirt, Yvirt;               /* more options */
long   delay;                                    /* even more options */
struct timespec rtime;
int    files_count = 0;
char **file_names = NULL;
double scalex, scaley;
int    graphcolor, fontheight;

#ifdef __LINUX__
int parse_options(int argc, char **argv);
#endif /* __LINUX__ */

int plot_file(const char *filename, int fileno);


int main(int argc, char **argv)
{
    char   c;
    int    i, fileno;

    progname = *argv;

    /* Default values */
    xvirt      = 0;
    Xvirt      = -1;  /* will be set to screen width */
    yvirt      = 0;
    Yvirt      = -1;  /* will be set to screen height */
    stretch    = 0;
    showaxes   = 1;
    axescolor  = AXES_COLOR;
    showlabels = 1;
    delay      = 0; /* 2000000000 */

#ifdef __LINUX__
    /* Process command line arguments */
    i = parse_options(argc, argv);
    if ( i == 0 )
        return -1;
    else if ( i == 1 )
        return 0;
#else /* __LINUX__ */
    file_names = argv + 1;  /* where the filenames begin */
    files_count = argc - 1; /* how many of them */
#endif /* __LINUX__ */

    /* Setup graphics */
    VGA_INIT;
    fontheight = 16;
    if ( !loadfont(8, fontheight, "font.16") ) {
        VGA_CLOSE;
        fprintf(stderr, errFont, *argv);
        return -1;
    }

    scrwidth   = vga_getxdim();
    scrheight  = vga_getydim();
    colordepth = vga_getcolors();

    /* Set viewport coordinates */
    if ( xvirt < 0 )
        xvirt = 0;
    if ( yvirt < 0 )
        yvirt = 0;
    if ( Xvirt < 0 )
        Xvirt = scrwidth - 1;
    if ( Yvirt < 0 )
        Yvirt = scrheight - 1;
    if ( Xvirt < xvirt )
        i = xvirt, xvirt = Xvirt, Xvirt = i;
    if ( Yvirt < yvirt )
        i = yvirt, yvirt = Yvirt, Yvirt = i;

    /* Set delay */
    rtime.tv_sec  = (time_t) delay / 1000000000.l;
    rtime.tv_nsec = (long) delay % 1000000000l; /* 100000 */

    /* Start reading and plotting the data files */
    for ( fileno = 0; fileno < files_count; fileno++ )
        if ( plot_file(*(file_names + fileno), fileno) )
            return -1;

    /* Wait for a key and bail out */
    scanf("%c", &c);
    VGA_CLOSE;

    return 0;
}


int plot_file(const char *filename, int fileno)
{
    char   c;
    FILE  *datafile;
    int    i, xcur, ycur, xlast, ylast;
    int    cont, bailout;
    double x, y;
    double xreal, Xreal, yreal, Yreal;
    double temp;

    /* Open next file */
    datafile = fopen( filename, "r" );
    if ( !datafile ) {
        VGA_CLOSE;
        fprintf(stderr, errOpenFile, progname, filename);
        return -1;
    }

    /* Read in the label */
    i = 0;
    while ( (c = fgetc(datafile)) != EOF && c != '\n' ) {
        *(label + i++) = c;
        if ( i == LABEL_LEN - 1) { /* No more space and label not finished */
            VGA_CLOSE;
            fprintf(stderr, errLabelTooLong,
                    progname, filename, LABEL_LEN-1);
            return -1;
        }
    }
    *(label + i) = '\0';

    /* Read in graph color and real-world viewport dimensions */
    i = fscanf(datafile, "%d%lf%lf%lf%lf",
               &graphcolor, &xreal, &yreal, &Xreal, &Yreal);
    if ( i == EOF ) {
        VGA_CLOSE;
        fprintf(stderr, errUnexpectedEOF, progname, filename);
        return -1;
    }
    if ( i != 5 ) {
        VGA_CLOSE;
        fprintf(stderr, errParsingFile, progname, filename);
        return -1;
    }

    /* Calculate scale factors and draw axes only if it is the first file */
    if ( fileno == 0 ) {
        if ( xreal > Xreal )
            temp = xreal, xreal = Xreal, Xreal = temp;
        if ( yreal > Yreal )
            temp = yreal, yreal = Yreal, Yreal = temp;
        /* Scale factors */
        scalex = (Xvirt - xvirt) / (Xreal - xreal);
        scaley = (Yvirt - yvirt) / (Yreal - yreal);
        if ( !stretch )
            if ( scalex < scaley )
                scaley = scalex;
            else
                scalex = scaley;
        /* Draw coordinate axes */
        if ( showaxes ) {
            vga_setcolor(axescolor);
            if ( xreal < 0 && Xreal > 0 ) {
                ycur = Yvirt + yreal * scaley; /* y = 0 */
                vga_drawline(xvirt, ycur, Xvirt, ycur);
            }
            if ( yreal < 0 && Yreal > 0 ) {
                xcur = xvirt - xreal * scalex; /* x = 0 */
                vga_drawline(xcur, yvirt, xcur, Yvirt);
            }
        }
        xlast = ylast = 0;
    }

    /* Read and plot the data with the requested color */
    vga_setcolor(graphcolor);
    bailout = 0;
    while ( !bailout ) {
        if ( EOF == fscanf(datafile, "%d%lf%lf", &cont, &x, &y) )
            bailout = 1;
        else {
            xcur = xvirt + (x - xreal) * scalex;
            ycur = Yvirt - (y - yreal) * scaley;
            if ( cont )
                vga_drawline(xlast, ylast, xcur, ycur);
            if ( delay )
                nanosleep(&rtime, NULL);
            xlast = xcur;
            ylast = ycur;
        }
    }
    /* Display label */
    if ( showlabels ) {
        vga_setcolor(graphcolor);
        textout(xvirt, yvirt + fileno * fontheight, label);
    }

    fclose(datafile);

    return 0;
}


#ifdef __LINUX__
/* Return codes:
 * 0    - error, exit(-1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
int parse_options(int argc, char **argv)
{
    char c, *ptr;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":x:X:y:Y:a:ALsd:h")) != -1 )
        switch ( c ) {
        case 'a':
            axescolor = strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "axes color");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            break;
        case 'x':
            xvirt = strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "xmin");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            break;
        case 'y':
            yvirt = strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "ymin");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            break;
        case 'X':
            Xvirt = strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "xmax");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            break;
        case 'Y':
            Yvirt = strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "ymax");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            break;
        case 'A':
            showaxes = 0;
            break;
        case 'L':
            showlabels = 0;
            break;
        case 's':
            stretch = 1;
            break;
        case 'd':
            delay = strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "delay");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            break;
        case 'h':
            printf(usage, *argv, AXES_COLOR);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, *argv, optopt);
                fprintf(stderr, usehelp, *argv);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, *argv, optopt);
                fprintf(stderr, usehelp, *argv);
            }
            return 0;
            break;
        case ':': /* missing option argument */
            switch ( optopt ) {
            case 'a':
                fprintf(stderr, errArgumentReq, *argv, "-a", "axes color");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            case 'x':
                fprintf(stderr, errArgumentReq, *argv, "-x", "xmin");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            case 'y':
                fprintf(stderr, errArgumentReq, *argv, "-y", "ymin");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            case 'X':
                fprintf(stderr, errArgumentReq, *argv, "-X", "xmax");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            case 'Y':
                fprintf(stderr, errArgumentReq, *argv, "-Y", "ymax");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            case 'd':
                fprintf(stderr, errArgumentReq, *argv, "-d", "delay");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            default:
                abort(); /* should never get here */
                break;
            }
        default:
            abort(); /* should never happen */
        }

    if ( optind == argc ) {
        fprintf(stderr, errDataFileReq, *argv);
        fprintf(stderr, usehelp, *argv);
        return 0;
    }

    file_names = argv + optind;  /* where the filenames begin */
    files_count = argc - optind; /* how many of them */

    return -1;
}
#endif /* __LINUX__ */



/*
 * Local variables:
 * compile-command: "make -k"
 * End:
 */
