#include <ctype.h>

#include "splines.h"

#include <time.h>


char      *progname;
GtkWidget *main_window;
GtkWidget *darea;


static char *errMemory   = "%s: error--cannot allocate memory\n";

static gint   darea_width   = 640;
static gint   darea_height  = 480;
static gint16 point_radius  = 5;
/* Take heed when increasing it--draw_splines() might overflow!!! */
static gint   granularity   = 50;
static gint   random_points = 7;

static pnode *points = NULL;
static pnode *dragging = NULL;


static int parse_options(int argc, char **argv);



static void picture_erase(GtkWidget *widget)
{
    draw_splines( (GdkDrawable *) widget->window,
                  main_window->style->bg_gc[GTK_STATE_NORMAL],
                  points, granularity );
    draw_points( (GdkDrawable *) widget->window,
                 main_window->style->bg_gc[GTK_STATE_NORMAL],
                 points, point_radius );

    return;
}

static void picture_draw(GtkWidget *widget)
{
    draw_splines( (GdkDrawable *) widget->window,
                  main_window->style->fg_gc[GTK_STATE_NORMAL],
                  points, granularity );
    draw_points( (GdkDrawable *) widget->window,
                 main_window->style->fg_gc[GTK_STATE_NORMAL],
                 points, point_radius );

    return;
}

/*
 * Signal and event handlers
 */
static void darea_destroy(GtkWidget *widget)
{
    return;
}

static gint darea_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
    picture_draw(widget);

    return FALSE;
}

static gint darea_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
    picture_draw(widget);

    return TRUE;
}


static gint darea_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
    GdkPoint p;
    GdkModifierType state;

    state = event->state;
    p.x = (gint16) event->x;
    p.y = (gint16) event->y;

    if ( event->button == 1 ) /* start dragging */
        dragging = point_from_coords(points, &p, point_radius);
    else if ( !dragging ) {    /* add or remove points */
        pnode *point;
        point = point_from_coords(points, &p, point_radius);
        picture_erase(widget);
        if ( point ) /* remove */
            point_remove( points, point );
        else {       /* add */
            if ( !(dragging = point_add(points, &p) ) ) {
                fprintf(stderr, errMemory, progname);
                gtk_exit(1);
            }
        }
        picture_draw(widget);
    }

    return TRUE;
}


static gint darea_motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
    GdkPoint p;
    gint x, y;
    GdkModifierType state;

    if ( !dragging )
        return FALSE;

    if ( event->is_hint ) {
        gdk_window_get_pointer( event->window, &x, &y, &state );
        p.x = (gint16) x;
        p.y = (gint16) y;
    }
    else {
        p.x = (gint16) event->x;
        p.y = (gint16) event->y;
        state = event->state;
    }

    picture_erase(widget);

    dragging->p = p;
    if ( dragging == points->next )          /* this is first point */
        points->p = p;         /* set also the first `invisible' point */
    else if ( dragging->next->next == NULL ) /* this is last point */
        dragging->next->p = p; /* set also the last `invisible' point */
    picture_draw(widget);

    return TRUE;
}


static gint darea_button_release_event(GtkWidget *widget, GdkEventButton *event)
{
    if ( dragging ) {
        GdkPoint p;
        GdkModifierType state;

        state = event->state;
        p.x = (gint16) event->x;
        p.y = (gint16) event->y;

        if ( p.x < 0 )
            p.x =  0;
        else if ( p.x > widget->allocation.width )
            p.x =  widget->allocation.width;

        if ( p.y < 0 )
            p.y =  0;
        else if ( p.y > widget->allocation.height )
            p.y =  widget->allocation.height;

        picture_erase(widget);
        dragging->p = p;
        picture_draw(widget);

        dragging = NULL;

        return TRUE;
    }

    return FALSE;
}



int main(int argc, char **argv)
{
    int i;
    GtkWidget *main_box;
    GdkPoint p;

    progname = *argv;

    /* GTK initialization */
    gtk_init(&argc, &argv);
    /* Process command line arguments */
    i = parse_options(argc, argv);
    if ( i == 0 )
        gtk_exit(1);
    else if ( i == 1 )
        gtk_exit(0);

    if ( !(points = points_init()) ) {
        fprintf(stderr, errMemory, progname);
        gtk_exit(1);
    }

    /* Generate random points */
    srand(time(NULL));
    for ( i = 0; i < random_points; i++ ) {
        p.x = rand() / (RAND_MAX/darea_width);
        p.y = rand() / (RAND_MAX/darea_height);
        if ( !point_add(points, &p) ) {
            fprintf(stderr, errMemory, progname);
            gtk_exit(1);
        }
    }

    /* Main window */
    main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title( GTK_WINDOW(main_window), "Splines" );
    /* Can shrink and grow */
    gtk_window_set_policy( GTK_WINDOW(main_window), TRUE, TRUE, FALSE );
    /* Will quit on destroy signal event */
    gtk_signal_connect( GTK_OBJECT(main_window), "destroy",
                        GTK_SIGNAL_FUNC(gtk_main_quit), NULL );
    /* Pack main vbox into main window (homogeneous, spacing) */
    main_box = gtk_vbox_new(FALSE, 0);
    gtk_container_add ( GTK_CONTAINER(main_window), main_box );
    /* Pack drawing area into main box */
    gtk_widget_realize(main_window);
    darea = gtk_drawing_area_new();
    gtk_widget_set_usize( darea, darea_width, darea_height );
    /* (..., expand, fill, padding)*/
    gtk_box_pack_start ( GTK_BOX(main_box), darea, TRUE, TRUE, 0 );
    gtk_signal_connect( GTK_OBJECT(darea), "expose_event",
                        GTK_SIGNAL_FUNC(darea_expose_event), NULL );
    gtk_signal_connect( GTK_OBJECT(darea), "configure_event",
                        GTK_SIGNAL_FUNC(darea_configure_event), NULL );
    gtk_signal_connect( GTK_OBJECT(darea), "button_press_event",
                        GTK_SIGNAL_FUNC(darea_button_press_event), NULL );
    gtk_signal_connect( GTK_OBJECT(darea), "button_release_event",
                        GTK_SIGNAL_FUNC(darea_button_release_event), NULL );
    gtk_signal_connect( GTK_OBJECT(darea), "motion_notify_event",
                        GTK_SIGNAL_FUNC(darea_motion_notify_event), NULL );
    gtk_signal_connect( GTK_OBJECT(darea), "destroy",
                        GTK_SIGNAL_FUNC(darea_destroy), NULL );
    gtk_widget_set_events(
        darea,
        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
        GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK );

    /* Pop up the whole shebang */
    gtk_widget_show(main_box);
    gtk_widget_show(darea);
    gtk_widget_show(main_window);

    gtk_main();

    return 0;
}



static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character \
`\\x%x'\n";
static char *errDAreaDimen    = "%s: error--drawing area dimensions must be positive integers\n";
static char *errPointRadius   = "%s: error--point radius must be non-negative integer\n";
static char *errGranularity   = "%s: error--calculation precision must be an integer in range [1, 100]\n";
static char *errNumPoints     = "%s: error--number of random points must be a non-negative integer\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument (%s)\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\n\
Usage: %s [-h] [-x width] [-y height] [-r radius] [-p precision] [-n number]\n\
\n\
-h              display this help message\n\
-x width        drawing area's width                           (default: %3d)\n\
-y height       drawing area's height                          (default: %3d)\n\
-r radius       point radius                                   (default: %3d)\n\
-p precision    splines calculation precision (range: [1, 100]; default: %3d)\n\
-n number       how many random points to generate on startup  (default: %3d)\n\
\n\
Note: left-click and drag any point to adjust the spline;
      right-click on existing point to delete it;\n\
      right-click on free space to add a new point to the end of list\n\
\n";

/* Return codes:
 * 0    - error, exit(1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
static int parse_options(int argc, char **argv)
{
    char c, *ptr;
    gint  i;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":x:y:r:p:n:h")) != -1 )
        switch ( c ) {
        case 'x':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 0 ) {
                fprintf(stderr, errDAreaDimen, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            darea_width = i;
            break;
        case 'y':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 0 ) {
                fprintf(stderr, errDAreaDimen, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            darea_height = i;
            break;
        case 'r':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 0 ) {
                fprintf(stderr, errPointRadius, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            point_radius = i;
            break;
        case 'p':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 1 || i > 100 ) {
                fprintf(stderr, errGranularity, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            granularity = i;
            break;
        case 'n':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i < 0 ) {
                fprintf(stderr, errNumPoints, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            random_points = i;
            break;
        case 'h':
            printf(usage, progname, darea_width, darea_height, point_radius,
                   granularity, random_points);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            return 0;
            break;
        case ':': { /* missing option argument */
            char *msg = NULL;

            switch ( optopt ) {
            case 'x':
                msg = "drawing area width";
                break;
            case 'y':
                msg = "drawing area height";
                break;
            case 'r':
                msg = "point radius";
                break;
            case 'p':
                msg = "calculation precision";
                break;
            case 'n':
                msg = "number of points";
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
            fprintf(stderr, errArgumentReq, progname, optopt, msg);
            fprintf(stderr, usehelp, progname);
            return 0;
            break;
        }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }

    return -1;
}



/*
 * Local variables:
 * compile-command: "make -k && ./splines"
 * End:
 */
