#include "triangulate.h"
#include <unistd.h>
#include <ctype.h>


static GtkWidget *main_window    = NULL;
static GtkWidget *drawing_area   = NULL;
static GdkPixmap *backing_pixmap = NULL;

static gint drawing_area_width  = DEFAULT_WIDTH;
static gint drawing_area_height = DEFAULT_HEIGHT;
static int  use_stdin = 1, use_stdout = 1, use_X = 0;
static int  GTK_initialized = 0;

static list *points = NULL, *triangles = NULL;
static gboolean drawing = FALSE;
static GdkPoint drawing_to;

static       char *progname = NULL;
static const char *errMemory    = "%s: error--cannot allocate memory\n";
static const char *errGtkInit   = "%s: error--cannot initialize GTK\n";
static const char *errPointList = "%s: dead cycle detected--point list does not describe a valid polygon\n";

static const char *msgNewPolygon = "%s: list of vertices of a new polygon\n";
static const char *msgTriangList = "%s: triangulation list for the new polygon\n";

static int parse_options(int argc, char **argv);



/* Draws a line from last point of points list to the passed point, both in
 * backing_pixmap and in drawing_area */
static void draw_transient_line(void)
{
    GdkGC *gc;

    /* Draw `inverted' line */
    gc = drawing_area->style->fg_gc[GTK_STATE_NORMAL];
    gdk_gc_set_function( gc, GDK_INVERT );
    ((pitem *) points->tail())->lineto(
        (GdkDrawable *) backing_pixmap, gc, drawing_to );
    ((pitem *) points->tail())->lineto(
        (GdkDrawable *) drawing_area->window, gc, drawing_to );
    /* Restore COPY function */
    gdk_gc_set_function( gc, GDK_COPY );

    return;
}

/* Draws triangles both in backing_pixmap and in drawing_area */
static void draw_triangles(void)
{
    if ( triangles ) {
        GdkGC *gc;
        gc = drawing_area->style->fg_gc[GTK_STATE_INSENSITIVE];
        triangles->draw( (GdkDrawable *) backing_pixmap, gc );
        triangles->draw( (GdkDrawable *) drawing_area->window, gc );
    }

    return;
}

/* Draws polygon both in backing_pixmap and in drawing_area.  If close == TRUE,
 * connects last point in the list with the first; otherwise does not */
static void draw_polygon(gboolean close)
{
    if ( points ) {
        pitem *prev, *curr;
        GdkGC *gc;

        if ( close ) { /* start by connecting last vertex with first */
            prev = (pitem *) points->tail();
            curr = (pitem *) points->head();
        } else {       /* make sure last vertex never connects to first */
            prev = (pitem *) points->head();
            if ( prev ) curr = (pitem *) prev->next; else curr = NULL;
        }
        gc = drawing_area->style->fg_gc[GTK_STATE_NORMAL];

        while ( curr ) {
            curr->lineto( (GdkDrawable *) backing_pixmap, gc, prev->p() );
            curr->lineto( (GdkDrawable *) drawing_area->window, gc, prev->p() );
            prev = curr;
            curr = (pitem *) curr->next;
        }
    }

    return;
}


/*
 * Signal and event handlers
 */
static void drawing_area_destroy( GtkWidget *widget )
{
    gdk_pixmap_unref(backing_pixmap);

    return;
}

static gint drawing_area_expose_event( GtkWidget *widget, GdkEventExpose *event )
{
    /* Redraw clobbered area of backing pixmap */
    gdk_draw_pixmap( widget->window,
                     widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                     backing_pixmap,
                     event->area.x, event->area.y,
                     event->area.x, event->area.y,
                     event->area.width, event->area.height );

    return FALSE;
}

/* In configure event we (re)create the backing pixmap with new size, hide
 * transient drawing line, and redraw polygon and triangles */
static gint drawing_area_configure_event( GtkWidget *widget,
                                          GdkEventConfigure *event )
{
    gint width, height;

    if ( backing_pixmap )
        gdk_pixmap_unref(backing_pixmap);

    width  = widget->allocation.width;
    height = widget->allocation.height;

    backing_pixmap = gdk_pixmap_new(widget->window, width, height, -1);

    /* Blank out the rectangle (only in backing_pixmap) */
    gdk_draw_rectangle( backing_pixmap,
                        widget->style->bg_gc[GTK_STATE_NORMAL],
                        TRUE, 0, 0, width, height );

    /* Set drawing_to to the last vertex in points list, so that next invocation
     * of motion_notify_event handler will draw nothing (a zero-length transient
     * line) */
    if ( drawing ) {
        assert(points);
        drawing_to = ((pitem *) points->tail())->p();
    }

    /* Draw triangles */
    draw_triangles();

    /* Draw polygon.  If polygon is completed (`!drawing'), connect last vertex
     * with first; otherwise don't */
    draw_polygon(!drawing);

    return TRUE;
}


static gint drawing_area_button_press_event( GtkWidget *widget,
                                             GdkEventButton *event )
{
    assert(backing_pixmap);

    if ( event->button == 1 ) { /* 1st mouse button--drawing */
        int x = (int) event->x, y = (int) event->y;
        pitem *curr;

        /* If we aren't drawing, user is starting drawing a new polygon */
        if ( !drawing ) {
            gint width, height;

            if ( points ) {
                /* Free old lists */
                delete points;
                points = NULL;
                if ( triangles ) {
                    delete triangles;
                    triangles = NULL;
                }
            }

            /* Allocate new points list */
            if ( !points )
                points = new list;
            if ( !points ) {
                fprintf(stderr, errMemory, progname);
                gtk_exit(1);
            }

            width  = widget->allocation.width;
            height = widget->allocation.height;

            /* Blank out backing store and window */
            gdk_draw_rectangle( backing_pixmap,
                                widget->style->bg_gc[GTK_STATE_NORMAL],
                                TRUE, 0, 0, width, height );
            gdk_draw_rectangle( (GdkDrawable *) widget->window,
                                widget->style->bg_gc[GTK_STATE_NORMAL],
                                TRUE, 0, 0, width, height );
        }

        assert(points);

        /* Make sure this point is not in the list */
        curr = (pitem *) points->head();
        while ( curr ) {
            if ( curr->p().x == x && curr->p().y == y )
                break;
            curr = (pitem *) curr->next;
        }
        if ( !curr ) { /* didn't find such point in the list */
            /* Connect last point in points list with this point, if
             * necessary */
            if ( drawing ) {
                /* Draw solid line in backing store and in the window */
                ((pitem *) points->tail())->lineto(
                    (GdkDrawable *) backing_pixmap,
                    widget->style->fg_gc[GTK_STATE_NORMAL],
                    x, y );
                ((pitem *) points->tail())->lineto(
                    (GdkDrawable *) drawing_area->window,
                    widget->style->fg_gc[GTK_STATE_NORMAL],
                    x, y );
            }
            /* Add this point to the list of vertices */
            if ( use_stdout && !points->count() ) /* this is a first vertex */
                fprintf(stdout, msgNewPolygon, progname);
            points->add( new pitem(x, y) );
            if ( use_stdout )
                points->tail()->print(stdout);
            /* Set drawing_to point to the new vertex, so that next invocation
             * of motion_notify_event handler will draw nothing (a zero-length
             * transient line) */
            drawing_to.x = (gint16) x;
            drawing_to.y = (gint16) y;
            drawing = TRUE;
        }
    } else { /* not a 1st mouse button--user finished drawing */
        /* Erase old transient line */
        if ( drawing )
            draw_transient_line();
        drawing = FALSE;
        /* Draw polygon and triangles */
        if ( triangles )
            delete triangles;
        triangles = triangulate(points);
        draw_triangles();
        draw_polygon(TRUE); /* connect last vertex with first */
        /* Print out the triangles list, if necessary */
        if ( use_stdout ) {
            if ( !triangles )
                fprintf(stdout, errPointList, progname);
            else {
                fprintf(stdout, msgTriangList, progname);
                triangles->print(stdout);
            }
            fputc('\n', stdout);
        }
    }

    return TRUE;
}

static gint drawing_area_motion_notify_event( GtkWidget *widget,
                                              GdkEventMotion *event )
{
    gint   x, y;
    GdkModifierType  state;

    if ( !drawing )
        return TRUE;

    assert(backing_pixmap);

    /* Get pointer coordinates */
    if ( event->is_hint )
        gdk_window_get_pointer (event->window, &x, &y, &state);
    else {
        x = (gint) event->x;
        y = (gint) event->y;
        state = (GdkModifierType) event->state;
    }

    /* Invert old line */
    draw_transient_line();
    /* Draw new inverted line */
    drawing_to.x = (gint16) x;
    drawing_to.y = (gint16) y;
    draw_transient_line();

    return TRUE;
}


int main(int argc, char **argv)
{
    int  i, x, y;

    /* GTK initialization--too bad cannot do it after parse_options() because
     * user can give GTK's options which parse_options() does not recognize */
    GTK_initialized = gtk_init_check(&argc, &argv);

    /* Process command line arguments */
    progname = *argv;
    i = parse_options(argc, argv);
    if ( i == 0 )
        return 1;
    else if ( i == 1 )
        return 0;

    /* Read in points, if requested */
    if ( use_stdin ) {
        points = new list;
        if ( !points ) {
            fprintf(stderr, errMemory, progname);
            return 1;
        }
        while ( 2 == scanf("%d %d", &x, &y) )
            points->add(new pitem(x, y));

        triangles = triangulate(points);
        if ( !triangles )
            fprintf(stderr, errPointList, progname);

        /* Print out the list, if necessary */
        if ( use_stdout ) {
            if ( triangles ) {
                fprintf(stdout, msgTriangList, progname);
                triangles->print(stdout);
                fputc('\n', stdout);
            }
        }
    }

    if ( use_X ) {
        if ( !GTK_initialized )
            fprintf(stderr, errGtkInit, progname);
        else {
            /* Main window */
            main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
            gtk_window_set_title( GTK_WINDOW(main_window),
                                  "Polygon Triangulation" );
            gtk_window_set_policy( GTK_WINDOW(main_window), TRUE, TRUE, FALSE );
            gtk_signal_connect( GTK_OBJECT(main_window), "destroy",
                                GTK_SIGNAL_FUNC(gtk_main_quit),
                                (gpointer) NULL );

            /* Drawing area */
            drawing_area = gtk_drawing_area_new();
            gtk_widget_set_usize( drawing_area,
                                  drawing_area_width, drawing_area_height );
            gtk_container_add( GTK_CONTAINER(main_window), drawing_area );
            gtk_signal_connect( GTK_OBJECT(drawing_area), "expose_event",
                                GTK_SIGNAL_FUNC(drawing_area_expose_event),
                                (gpointer) NULL );
            gtk_signal_connect( GTK_OBJECT(drawing_area), "configure_event",
                                GTK_SIGNAL_FUNC(drawing_area_configure_event),
                                (gpointer) NULL );
            gtk_signal_connect( GTK_OBJECT(drawing_area), "destroy",
                                GTK_SIGNAL_FUNC(drawing_area_destroy),
                                (gpointer) NULL );
            gtk_signal_connect( GTK_OBJECT(drawing_area), "button_press_event",
                                GTK_SIGNAL_FUNC(drawing_area_button_press_event),
                                (gpointer) NULL );
            gtk_signal_connect( GTK_OBJECT (drawing_area), "motion_notify_event",
                                GTK_SIGNAL_FUNC(drawing_area_motion_notify_event),
                                (gpointer) NULL);

            gtk_widget_set_events( drawing_area,
                                   GDK_EXPOSURE_MASK |
                                   GDK_BUTTON_PRESS_MASK |
                                   GDK_POINTER_MOTION_MASK |
                                   GDK_POINTER_MOTION_HINT_MASK);

            gtk_widget_show(drawing_area);
            gtk_widget_show(main_window);

            gtk_main();
        }
    }

    return 0;
}



static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
static char *errDimensions    = "%s: error--drawing area dimensions must be a positive integer\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument (%s)\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\n\
Usage: %s [-h] [-i | -I] [-o | -O] [-d | -D] [-x width] [-y height]\n\
\n\
-h   display this help message\n\
-i   force input from console (read list of polygon vertices from stdin)\n\
-I   suppress input from console\n\
-o   force output to console (print list of points/triangles to stdout)\n\
-O   suppress output to console\n\
-d   draw polygon and triangles in an X window (requires GTK)\n\
-D   do not draw polygon and triangles in an X window\n\
-x   width of drawing area in the X window  (default:%4d)\n\
-y   height of drawing area in the X window (default:%4d)\n\
\n\
Notes:  1. By default, list of polygon vertices (pairs of integers) is read\n\
           from stdin, list of trianlegs (triplets of points) is output to\n\
           stdout; no X window is created (i.e., options `-ioD' are assumed)\n\
        2. When feeding list of polygon vertices to stdin, arrange the points\n\
           clockwise, and do not close last point with first (do not duplicate\n\
           first vertex at the end of the list)\n\
        3. When drawing in an X window is selected, you can draw a polygon in\n\
           the X window with a mouse.  Left-click to place a vertex of the\n\
           polygon under mouse pointer; when done drawing, right-click your\n\
           mouse to kick off the polygon's triangulation.  List of the\n\
           vertices will be output to console (if not suppressed by `-O'\n\
           option), resulting triangles will be drawn in the window and also\n\
           output to console (if not suppressed by `-O' option)\n\
\n";

/* Return codes:
 * 0    - error, exit(1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
static int parse_options(int argc, char **argv)
{
    char c, *ptr;
    gint i;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":iIoOdDx:y:h")) != -1 )
        switch ( c ) {
        case 'i':
            use_stdin = 1;
            break;
        case 'I':
            use_stdin = 0;
            break;
        case 'o':
            use_stdout = 1;
            break;
        case 'O':
            use_stdout = 0;
            break;
        case 'd':
            use_X = 1;
            break;
        case 'D':
            use_X = 0;
            break;
        case 'x':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i <= 0 ) {
                fprintf(stderr, errDimensions, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            drawing_area_width = i;
            break;
        case 'y':
            i = (gint) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i <= 0 ) {
                fprintf(stderr, errDimensions, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            drawing_area_height = i;
            break;
        case 'h':
            printf(usage, progname, DEFAULT_WIDTH, DEFAULT_HEIGHT);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            return 0;
            break;
        case ':': { /* missing option argument */
            char *msg = NULL;

            switch ( optopt ) {
            case 'x':
                msg = "drawing area width";
                break;
            case 'y':
                msg = "drawing area height";
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
            fprintf(stderr, errArgumentReq, progname, optopt, msg);
            fprintf(stderr, usehelp, progname);
            return 0;
            break;
        }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }

    return -1;
}



/*
 * Local variables:
 * compile-command: "make -k && ./triangulate -d << EOF
100 200
500 300
200 400
300 300
150 250
EOF"
 * End:
 */
