#include "triangulate.h"

/*
 * Takes three GdkPoint's p0, p1, and p2.
 *
 * Returns:
 *
 *   positive number if points p0, p1, and p2 are oriented (point p1 lies left
 *   of line p0--p2);
 *
 *   negative number if points p0, p1, and p2 are disoriented (point p1 lies
 *   right of line p0--p2);
 *
 *   xaero if points p0, p1, and p2 lie on a line
 */
#define ORIENTED(p0, p1, p2)    ( (p1.x - p0.x)*(p2.y - p0.y) - \
                                  (p2.x - p0.x)*(p1.y - p0.y)   )

/*
 * Triangulates polygon described by points in points list.  Points must be
 * listed clockwise.  Last point in the list should not close with first,
 * i.e. last point should not be duplicated at the end of the list.
 *
 * Returns list of triangles on success, NULL on error (points list does not
 * describe a valid polygon)
 */
list *triangulate(list *points)
{
    list  *triangles;
    list  *vertices;
    pitem *p0, *p1, *p2, *p;
    int    orient0, orient1, orient2;
    int    cycles;

    if ( !points || points->count() < 3 ) /* nothing to triangulate */
        return NULL;

    triangles = new list;

    /* Make a scrap list of vertices and take first three points */
    vertices = points->copy();
    p0 = (pitem *) vertices->head();
    p1 = (pitem *) p0->next;
    p2 = (pitem *) p1->next;
    /* Cycles count */
    cycles = vertices->count();

    /* We could just go on `while ( vertices->count() > 3 )' and then just add
     * the remaining three points to triangles list, but passing the last three
     * points thru the loop will assure that they are an oriented tripple */
    while ( vertices->count() >= 3 ) {
        /* Check that p0, p1, and p2 are an oriented tripple (p1 lies left of
         * vector p0--p2) */
        if ( ORIENTED(p0->p(), p1->p(), p2->p()) > 0 ) {
            /* Now make sure that no vertex of the polygon lies inside of or on
             * a side of the triangle p0--p1--p2 */
            p = (pitem *) points->head();
            while ( p ) {
                /* Skip vertices which are points p0, p1, or p2 */
                if ( (p->p().x != p0->p().x || p->p().y != p0->p().y) &&
                     (p->p().x != p1->p().x || p->p().y != p1->p().y) &&
                     (p->p().x != p2->p().x || p->p().y != p2->p().y ) ) {
                    orient0 = ORIENTED(p0->p(), p->p(), p1->p());
                    orient1 = ORIENTED(p1->p(), p->p(), p2->p());
                    orient2 = ORIENTED(p2->p(), p->p(), p0->p());
                    /* Point lies inside a triangle if it lies on the left of
                     * each of the sides of the triangle, or on the right of
                     * each of the sides of the triangle.  We are also checking
                     * that point doesn't lie on triangle's side (when either
                     * two of the orient*'s are zero, or one is zero and other
                     * two have same sign) */
                    /* A point cannot lie left of all sides of the
                     * triangle--points are looked up clockwise */
//                     if ( orient0 >= 0 && orient1 >= 0 && orient2 >= 0 )
//                         break;
                    if ( orient0 <= 0 && orient1 <= 0 && orient2 <= 0 )
                        break;
                }
                p = (pitem *) p->next; /* move on */
            }
            if ( !p ) {  /* violating point not found, add a new triangle */
                triangles->add(new titem(p0->p(), p1->p(), p2->p()));
                vertices->del((pitem *) p1);    /* remove middle point */
                cycles = 1 + vertices->count(); /* restart cycles count */
                p1 = p0; /* so the code below would work */
            }
        }
        /* If we went around the vertices list without adding any triangles,
         * this means we're having a dead cycle (user gave invalid point list or
         * this code is broken) */
        if ( !cycles ) { /* dead cycle detected */
            delete triangles;
            triangles = NULL;
            break;
        }
        /* Update cycles count */
        cycles--;
        /* Move on to next three points */
        p0 = p1;
        p1 = p2;
        p2 = (pitem *) vertices->next_cycled(p2);
    }

    delete vertices; /* this was a scrap list */

    return triangles;
}



/*
 * Local variables:
 * compile-command: "make -k && ./triangulate < polygon.pts"
 * End:
 */
